import sys
import types
import base64
import ast


def _mock(name):
    class Mock(object):
        def __init__(self, *args, **kwargs):
            self.name = name
            self.args = args
            self.kwargs = kwargs

        def __eq__(self, other):
            if isinstance(other, Mock):
                return (
                    self.name == other.name
                    and self.args == other.args
                    and self.kwargs == other.kwargs
                )
            return False

    return Mock


class NoBSON(types.ModuleType):
    import re
    import datetime
    import calendar
    from collections import OrderedDict
    from json import (
        JSONEncoder,
    )

    class BSONError(Exception):
        """Base class for all BSON exceptions."""

    class InvalidId(BSONError):
        """Raised when trying to create an ObjectId from invalid data."""

    class InvalidDocument(BSONError):
        """Raised when trying to create a BSON object from an invalid document."""

    SON = _mock("SON")
    BSON = _mock("BSON")
    Code = _mock("Code")
    Int64 = _mock("Int64")
    Regex = _mock("Regex")
    MinKey = _mock("MinKey")
    MaxKey = _mock("MaxKey")
    Binary = _mock("Binary")
    Timestamp = _mock("Timestamp")
    Decimal128 = _mock("Decimal128")
    RawBSONDocument = _mock("RawBSONDocument")

    decimal128_NaN = Decimal128("NaN")
    decimal128_INF = Decimal128("Infinity")
    decimal128_NaN_ls = (
        decimal128_NaN,
        Decimal128("-NaN"),
        Decimal128("sNaN"),
        Decimal128("-sNaN"),
    )

    from .objectid import ObjectId
    from .tz_util import utc

    class CodecOptions(object):
        def __init__(self, document_class=dict, tz_aware=False, tzinfo=None):
            self.document_class = document_class
            self.tz_aware = tz_aware
            self.tzinfo = tzinfo

    DEFAULT_CODEC_OPTIONS = CodecOptions()

    def __init__(self):
        self.bson_used = False
        super(NoBSON, self).__init__(__name__)

    @classmethod
    def parse_codec_options(cls, options):
        default_opt = cls.DEFAULT_CODEC_OPTIONS
        return cls.CodecOptions(
            document_class=options.get("document_class", dict),
            tz_aware=options.get("tz_aware", default_opt.tz_aware),
            tzinfo=options.get("tzinfo", default_opt.tzinfo),
        )

    EPOCH_AWARE = datetime.datetime.fromtimestamp(0, utc)
    EPOCH_NAIVE = datetime.datetime.utcfromtimestamp(0)

    @classmethod
    def _datetime_to_millis(cls, dtm):
        """Convert datetime to milliseconds since epoch UTC."""
        if dtm.utcoffset() is not None:
            dtm = dtm - dtm.utcoffset()
        return int(
            cls.calendar.timegm(dtm.timetuple()) * 1000 + dtm.microsecond // 1000
        )

    @classmethod
    def _millis_to_datetime(cls, millis, opts):
        """Convert milliseconds since epoch UTC to datetime."""
        diff = ((millis % 1000) + 1000) % 1000
        seconds = (millis - diff) // 1000
        micros = diff * 1000
        if opts.tz_aware:
            dt = cls.EPOCH_AWARE + cls.datetime.timedelta(
                seconds=seconds, microseconds=micros
            )
            if opts.tzinfo:
                dt = dt.astimezone(opts.tzinfo)
            return dt
        else:
            return cls.EPOCH_NAIVE + cls.datetime.timedelta(
                seconds=seconds, microseconds=micros
            )

    class BSONEncoder(JSONEncoder):
        key_is_keyword = False

        def default(self, obj):
            self.key_is_keyword = True

            if isinstance(obj, NoBSON.ObjectId):
                return {"$oid": str(obj)}

            if isinstance(obj, bytes):
                return {"$binary": {"base64": base64.b64encode(obj).decode(),
                                    "subType": "00"}}

            if isinstance(obj, NoBSON.datetime.datetime):
                millis = NoBSON._datetime_to_millis(obj)
                return {"$date": millis}

            if isinstance(obj, NoBSON._re_type):
                flags = NoBSON._re_int_flag_to_str(obj.flags)
                if isinstance(obj.pattern, NoBSON._text_type):
                    pattern = obj.pattern
                else:
                    pattern = obj.pattern.decode("utf-8")
                return NoBSON.OrderedDict([("$regex", pattern), ("$options", flags)])

            if hasattr(obj, "to_json"):
                return obj.to_json()

            return NoBSON.JSONEncoder.default(self, obj)

    @classmethod
    def _key_validate(cls, key):
        if "." in key:
            raise cls.InvalidDocument("key '%s' must not contain '.'" % key)
        if key.startswith("$"):
            raise cls.InvalidDocument("key '%s' must not start with '$'" % key)

    @classmethod
    def document_encode(cls, doc, check_keys=False, *args, **kwargs):
        _encoder = cls.BSONEncoder()
        _key_sep = _encoder.key_separator
        candidate = serialized = ""
        for s in _encoder.iterencode(doc):
            if s == _key_sep:
                if not _encoder.key_is_keyword:
                    key = ast.literal_eval(candidate)
                    if not isinstance(key, cls._string_types):
                        raise cls.InvalidDocument(
                            "documents must have only string keys, key was %r" % key
                        )
                    if check_keys:
                        cls._key_validate(key)
                _encoder.key_is_keyword = False
            candidate = s
            serialized += s
        return serialized.encode()

    @classmethod
    def object_hook(cls, obj, opts=DEFAULT_CODEC_OPTIONS):
        if "$oid" in obj:
            return cls.ObjectId(obj["$oid"])
        if "$binary" in obj:
            return base64.b64decode(obj["$binary"]["base64"])
        if "$date" in obj:
            return cls._millis_to_datetime(int(obj["$date"]), opts)
        if "$regex" in obj:
            flags = cls._re_str_flags_to_int(obj.get("$options", ""))
            return cls.re.compile(obj["$regex"], flags)
        for key in cls.custom_json_hooks:
            if key in obj:
                return cls.custom_json_hooks[key](obj, opts)
        return obj

    @classmethod
    def document_decode(cls, serialized, codec_options=None, *args, **kwargs):
        from json import loads as _loads

        opts = codec_options or cls.DEFAULT_CODEC_OPTIONS
        dcls = opts.document_class
        return _loads(
            serialized,
            object_pairs_hook=lambda pairs: cls.object_hook(dcls(pairs), opts),
        )

    @classmethod
    def json_loads(cls, serialized, *args, **kwarg):
        from json import loads as _loads

        return _loads(serialized, object_hook=cls.object_hook)

    @classmethod
    def json_dumps(cls, doc, *args, **kwarg):
        from json import dumps as _dumps

        return _dumps(doc, default=cls.BSONEncoder().default)

    @classmethod
    def id_encode(cls, id, *args, **kwargs):
        return cls.document_encode(id)

    custom_json_hooks = {}

    _re_type = type(re.compile(""))
    _re_flags = (
        ("i", re.IGNORECASE),
        ("l", re.LOCALE),
        ("m", re.MULTILINE),
        ("s", re.DOTALL),
        ("u", re.UNICODE),
        ("x", re.VERBOSE),
    )

    if sys.version_info[0] == 3:
        _string_types = (str,)
        _integer_types = (int,)
        _text_type = str
    else:
        _string_types = (basestring,)  # noqa
        _integer_types = (int, long)  # noqa
        _text_type = unicode  # noqa

    @classmethod
    def _re_int_flag_to_str(cls, int_flags):
        return "".join(str_f for str_f, re_F in cls._re_flags if int_flags & re_F)

    @classmethod
    def _re_str_flags_to_int(cls, str_flags):
        flags = 0
        for str_f, re_F in cls._re_flags:
            flags |= re_F if str_f in str_flags else 0
        return flags



from __future__ import absolute_import
import types


class BSON_(types.ModuleType):

    from bson import (
        SON,
        BSON,
        Code,
        Int64,
        Regex,
        Binary,
        MinKey,
        MaxKey,
        ObjectId,
        Timestamp,
        Decimal128,
        _ENCODERS,
    )
    from bson.raw_bson import (
        RawBSONDocument,
    )
    from bson.json_util import (
        CANONICAL_JSON_OPTIONS,
    )
    from bson.codec_options import (
        CodecOptions,
        DEFAULT_CODEC_OPTIONS,
    )
    from bson.errors import (
        BSONError,
        InvalidId,
        InvalidDocument,
    )

    decimal128_NaN = Decimal128("NaN")
    decimal128_INF = Decimal128("Infinity")
    decimal128_NaN_ls = (
        decimal128_NaN,
        Decimal128("-NaN"),
        Decimal128("sNaN"),
        Decimal128("-sNaN"),
    )

    def __init__(self):
        self.bson_used = True
        super(BSON_, self).__init__(__name__)

    @classmethod
    def parse_codec_options(cls, options):
        from bson.codec_options import _parse_codec_options

        return _parse_codec_options(options)

    @classmethod
    def document_encode(
        cls, doc, check_keys=False, codec_options=DEFAULT_CODEC_OPTIONS
    ):
        return cls.BSON.encode(doc, check_keys=check_keys, codec_options=codec_options)

    @classmethod
    def document_decode(cls, doc, codec_options=DEFAULT_CODEC_OPTIONS):
        return cls.BSON(doc).decode(codec_options)

    @classmethod
    def json_loads(cls, serialized, json_options=None):
        from bson.json_util import loads as _loads

        json_options = json_options or cls.CANONICAL_JSON_OPTIONS
        return _loads(serialized, json_options=json_options)

    @classmethod
    def json_dumps(cls, doc, json_options=None):
        from bson.json_util import dumps as _dumps

        json_options = json_options or cls.CANONICAL_JSON_OPTIONS
        return _dumps(doc, json_options=json_options)

    @classmethod
    def id_encode(cls, id, codec_options=DEFAULT_CODEC_OPTIONS):
        # args: name, value, check_keys, opts
        return cls._ENCODERS[type(id)](b"", id, False, codec_options)



from __future__ import absolute_import
import sys


__all__ = [  # noqa: F822
    "SON",  # noqa: F822
    "BSON",  # noqa: F822
    "ObjectId",  # noqa: F822
    "Timestamp",  # noqa: F822
    "MinKey",  # noqa: F822
    "MaxKey",  # noqa: F822
    "Int64",  # noqa: F822
    "Decimal128",  # noqa: F822
    "Binary",  # noqa: F822
    "Regex",  # noqa: F822
    "Code",  # noqa: F822
    "RawBSONDocument",  # noqa: F822
    "CodecOptions",  # noqa: F822

    "decimal128_NaN",  # noqa: F822
    "decimal128_INF",  # noqa: F822
    "decimal128_NaN_ls",  # noqa: F822

    "BSONError",  # noqa: F822
    "InvalidId",  # noqa: F822
    "InvalidDocument",  # noqa: F822

    "id_encode",  # noqa: F822
    "document_encode",  # noqa: F822
    "document_decode",  # noqa: F822
    "json_loads",  # noqa: F822
    "json_dumps",  # noqa: F822
    "parse_codec_options",  # noqa: F822
]


bson_used = None


def __getattr__(name):
    if name in __all__:
        from .. import configure

        config = configure.session_config()
        init(config.get("use_bson"))
        return getattr(sys.modules[__name__], name)


def init(use_bson=None):
    from .. import errors

    self = sys.modules[__name__]
    if self.bson_used is not None:
        return

    # Init

    if use_bson is None:
        try:
            import bson  # noqa: F401
        except ImportError:
            use_bson = False
        else:
            use_bson = True

    if use_bson:
        from ._bson import BSON_
        bson_ = BSON_()
    else:
        from ._nobson import NoBSON
        bson_ = NoBSON()

    self.bson_used = bson_.bson_used
    for name in __all__:
        setattr(self, name, getattr(bson_, name))

    errors.init_bson_err()



import re
from copy import deepcopy
from datetime import datetime
from collections.abc import Mapping

from ..errors import OperationFailure

from .field_walker import FieldWalker
from .weighted import (
    Weighted,
    gravity,
    _cmp_decimal,
)
from ..types import (
    bson,
    RE_PATTERN_TYPE,
    integer_types,
    string_types,
    is_duckument_type,
    is_integer_type,
    is_pattern_type,
    keep,
    compare_documents,
    re_str_flags_to_int,
)


def validate_sort_specifier(sort):
    if not (is_integer_type(sort) and sort in (1, -1)):
        raise OperationFailure("bad sort specification", code=2)


def ordering(fieldwalkers, order, doc_type=None):
    """ """
    total = len(fieldwalkers)
    pre_sect_stack = []

    for path, revr in order.items():
        validate_sort_specifier(revr)

        is_reverse = bool(1 - revr)
        value_stack = []

        for indx, fieldwalker in enumerate(fieldwalkers):
            # get field value
            fieldwalker = FieldWalker(fieldwalker.doc, doc_type).go(path).get()
            values = list(fieldwalker.value.iter_flat())
            if values:
                value = tuple([Weighted(val) for val in values])

                if len(value):
                    # list will firstly compare with other doc by it's smallest
                    # or largest member
                    value = max(value) if is_reverse else min(value)

            elif not fieldwalker.value.is_exists():
                value = Weighted(None)

            else:
                # [] less than None
                value = (0, ())

            # read previous section
            pre_sect = pre_sect_stack[indx] if pre_sect_stack else 0
            # inverse if in reverse mode
            pre_sect = (total - pre_sect) if is_reverse else pre_sect
            indx = (total - indx) if is_reverse else indx

            value_stack.append((pre_sect, value, indx))

        # sort docs
        value_stack.sort(reverse=is_reverse)

        ordereddoc = []
        sect_stack = []
        sect_id = -1
        last_doc = None
        for _, value, indx in value_stack:
            # restore if in reverse mode
            indx = (total - indx) if is_reverse else indx
            ordereddoc.append(fieldwalkers[indx])

            # define section
            # maintain the sorting result in next level sorting
            if not value == last_doc:
                sect_id += 1
            sect_stack.append(sect_id)
            last_doc = value

        # save result for next level sorting
        fieldwalkers = ordereddoc
        pre_sect_stack = sect_stack

    return fieldwalkers


class LogicBox(list):
    """A callable operator/logic array for document filtering

    By defining a `theme`, the instance's bool processing logic will behave
    differently when being called, and return different operation result.

    Only operator functions or `LogicBox` type instance are valid member of
    the `LogicBox`.

    Args:
        theme (str): A document field path or logic name ($and/$or/$nor/$not).

    """

    def __init__(self, theme, implicitly=False):
        self.theme = theme
        self.implicitly = implicitly
        self._logic = {
            "$and": self._call_and,
            "$or": self._call_or,
            "$nor": self._call_nor,
            "$not": self._call_not,
            "$elemMatch": self._call_elemMatch,
        }

    @property
    def __name__(self):
        return self.__repr__()

    def __repr__(self):
        """Display `theme` and `LogicBox` or operators content within"""
        content = []
        name = "[{}]"
        if not self.implicitly:
            content = ["LogicBox({})".format(self.theme)]
            name = "{}"

        for i in self[:]:
            if callable(i):
                if hasattr(i, "_keep"):
                    # query ops
                    content.append("{}({})".format(i.__name__, i._keep()))
                else:
                    # LogicBox
                    content.append(i.__name__)
            else:
                content.append(i)

        return name.format(content)[1:-1].replace("'", "")

    def __call__(self, fieldwalker):
        """Recursively calling `LogicBox` or operators content within

        A short-circuit logic sub-structure, passing `FieldWalker` instance.

        Args:
            fieldwalker (FieldWalker): Received from `QueryFilter` instance.

        """
        try:
            return self._logic[self.theme](fieldwalker)
        except KeyError:
            return self._call_field(fieldwalker)

    def _gen(self, fieldwalker):
        return (cond(fieldwalker) for cond in self[:])

    def _call_field(self, fieldwalker):
        """Entering document field context before process"""
        with fieldwalker.go(self.theme).get():
            return all(self._gen(fieldwalker))

    def _call_elemMatch(self, fieldwalker):
        """ """
        with fieldwalker.value as field_value:
            for elem in field_value.iter_elements():

                def elem_iter():
                    try:
                        for e in elem:
                            yield e
                    except TypeError:
                        pass
                    finally:
                        yield elem

                field_value.change_iter(elem_iter)
                if all(self._gen(fieldwalker)):
                    return True

    def _call_and(self, fieldwalker):
        return all(self._gen(fieldwalker))

    def _call_or(self, fieldwalker):
        return any(self._gen(fieldwalker))

    def _call_nor(self, fieldwalker):
        return not any(self._gen(fieldwalker))

    def _call_not(self, fieldwalker):
        return not all(self._gen(fieldwalker))


class QueryFilter(object):
    """Document query filter

    Parsing MongoDB document query language, generate a callable instance for
    documents query filtering.

    Args:
        spec (dict): MongoDB document query language object.

    """

    def __init__(self, spec):

        # Top-level operators, work on top of fields.
        self.pathless_ops = {
            # Logical
            "$and": self.parse_logic("$and"),
            "$nor": self.parse_logic("$nor"),
            "$or": self.parse_logic("$or"),
            # Evaluation
            "$expr": None,
        }

        # Field-level operators, need to work inside a field context.
        self.field_ops = {
            # Logical
            "$not": self._parse_not,
            # Comparison
            "$eq": parse_eq,
            "$gt": parse_gt,
            "$gte": parse_gte,
            "$in": parse_in,
            "$lt": parse_lt,
            "$lte": parse_lte,
            "$ne": parse_ne,
            "$nin": parse_nin,
            # Element
            "$exists": parse_exists,
            "$type": parse_type,
            # Array
            "$all": parse_all,
            "$elemMatch": self._parse_elemMatch,
            "$size": parse_size,
            # Evaluation
            "$jsonSchema": parse_jsonSchema,
            "$mod": parse_mod,
            "$regex": parse_regex,
        }

        # Start parsing query object
        self.conditions = self.parser(spec)
        self.__fieldwalker = None

        # ready to be called.

    def __repr__(self):
        return "QueryFilter({})".format(str(self.conditions))

    def __call__(self, doc, doc_type=None):
        """Recursively calling `LogicBox` or operators content within

        A short-circuit logic structure to determine the document can pass the
        filter or not.

        Args:
            doc (dict): Document received from database.

        """
        self.__fieldwalker = FieldWalker(doc, doc_type)
        return all(cond(self.__fieldwalker) for cond in self.conditions)

    @property
    def fieldwalker(self):
        return self.__fieldwalker

    def parser(self, spec):
        """Top-level parser"""

        # Implementation of implicitly $and operation, fundamental query
        # container.
        logic_box = LogicBox("$and", implicitly=True)

        for path, sub_spec in spec.items():
            if path.startswith("$"):
                try:
                    logic_box.append(self.pathless_ops[path](sub_spec))
                except KeyError:
                    raise OperationFailure(
                        "unknown top level operator: {}".format(path)
                    )
            else:
                logic_box.append(self.subparser(path, sub_spec))

        return logic_box

    def subparser(self, path, sub_spec):
        """Field-level parser"""

        # Implementation of field-level operation container.
        logic_box = LogicBox(path)

        # There are two processing direction in field-level, one is filtering
        # with operators, the other is implicitly value $eq operation.
        # The direction was first defined by the expression value type, if is
        # <dict>, then by *first* key is starts with "$" or not.
        #
        # Example:
        #
        #   {"field.name": {"$ne": 5, "$exists": True}} -> by operators
        #   {"field.name": {"data": 5, "id": 2}}        -> value matching ($eq)
        #
        #   But if something like this, mixing operator and non-operator key.
        #   {"field.name": {"$eq": 5, "id": 2}}
        #
        #   Depend on which key get iter *first*, then this query might:
        #   1) return no document and without any error, or
        #   2) raise an "OperationFailure: unknown operator" error.
        #
        if isinstance(sub_spec, bson.Regex):
            sub_spec = {"$regex": sub_spec}

        if _is_expression_obj(sub_spec):
            # Modify `sub_spec` for $regex and $options
            # before parse to `logic_box`
            if "$regex" in sub_spec:
                sub_spec = _modify_regex_optins(sub_spec)
            elif "$options" in sub_spec:
                raise OperationFailure("$options needs a $regex")

            for op, value in sub_spec.items():
                # Regex can't do $ne directly
                if op == "$ne" and isinstance(value, (RE_PATTERN_TYPE, bson.Regex)):
                    raise OperationFailure("Can't have RegEx as arg to $ne.")
                # is predictable ?
                if op in ("$gt", "$gte", "$lt", "$lte"):
                    if isinstance(value, (RE_PATTERN_TYPE, bson.Regex)):
                        raise OperationFailure(
                            "Can't have RegEx as arg to predicate over "
                            "field {!r}.".format(path)
                        )

                try:
                    logic_box.append(self.field_ops[op](value))
                except KeyError:
                    raise OperationFailure("unknown operator: {}".format(op))
        else:
            logic_box.append(parse_eq(sub_spec))

        return logic_box

    def parse_logic(self, theme):
        """Logical operator parser (un-themed)"""

        def _parse_logic(sub_spec):
            """Themed logical operator"""
            if not isinstance(sub_spec, list):
                raise OperationFailure("{} must be an array".format(theme))

            logic_box = LogicBox(theme)

            for cond in sub_spec:
                if not is_duckument_type(cond):
                    raise OperationFailure(
                        "$or/$and/$nor entries need to be full objects"
                    )

                logic_box.append(self.parser(cond))
            return logic_box

        return _parse_logic

    def _parse_not(self, sub_spec):
        # $not logic only available in field-level
        if isinstance(sub_spec, (RE_PATTERN_TYPE, bson.Regex)):
            return self.subparser("$not", {"$regex": sub_spec})

        elif is_duckument_type(sub_spec):
            for op in sub_spec:
                if op not in self.field_ops:
                    raise OperationFailure("unknown operator: {}".format(op))
                if op == "$regex":
                    raise OperationFailure("$not cannot have a regex")

            return self.subparser("$not", sub_spec)

        else:
            raise OperationFailure("$not needs a regex or a document")

    def _parse_elemMatch(self, sub_spec):
        # $elemMatch only available in field-level
        if not is_duckument_type(sub_spec):
            raise OperationFailure("$elemMatch needs an Object")

        for op in sub_spec:
            if op in self.field_ops:
                return self.subparser("$elemMatch", sub_spec)
            elif not op.startswith("$") or op in self.pathless_ops:
                return parse_elemMatch(sub_spec)
            else:
                raise OperationFailure("unknown operator: {}".format(op))


def _is_expression_obj(sub_spec):
    return is_duckument_type(sub_spec) and next(iter(sub_spec)).startswith("$")


# Only for preserving `int` type flags to bypass
# internal "flags must be string" type check
class _FALG(object):
    def __init__(self, int_flags):
        self.retrieve = int_flags

    __slots__ = ("retrieve",)


def _regex_options_(regex_flag, opt_flag):
    pass


def _regex_options_v42(regex_flag, opt_flag):
    if regex_flag and opt_flag:
        raise OperationFailure("options set in both $regex and $options")


_regex_options_check = _regex_options_v42


def _modify_regex_optins(sub_spec):
    """Merging $regex and $options values in query document

    Besides string type value, field $regex accept `bson.Regex` and
    `re._pattern_type` in pymongo. For re.flags and $options, based
    on the key order of dict, seconded will override the first, if
    they both exists in the query document.
    """
    new_sub_spec = None
    _re = None
    regex_flags = ""
    opt_flags = ""
    flags = ""

    for key, val in sub_spec.items():
        if key == "$options":
            opt_flags = val
            flags = opt_flags
        if key == "$regex" and isinstance(val, (RE_PATTERN_TYPE, bson.Regex)):
            regex_flags = _FALG(val.flags)
            flags = regex_flags
            # We will deepcopy `sub_spec` later for merging "$regex" and
            # "$options" to query parser, but we can't deepcopy regex
            # object, so move it to somewhere else and retrieve it later.
            _re = sub_spec["$regex"]
            sub_spec["$regex"] = None

    _regex_options_check(regex_flags, opt_flags)

    new_sub_spec = deepcopy(sub_spec)
    new_sub_spec["$regex"] = {
        "pattern": _re.pattern if _re else sub_spec["$regex"],
        "flags": flags,
    }

    # (monument): This is edge case, and only MongoDB 4.0 don't fail the
    #   operation.
    #
    # if (MONTY_MONGO_COMPAT_36
    #         and "#" in new_sub_spec["$regex"]["pattern"].rsplit("\n")[-1]):
    #     # (NOTE) davidlatwe:
    #     #   if pound(#) char exists in $regex string value and not ends with
    #     #   newline(\n), Mongo raise error. (but the message seems incomplete)
    #     raise OperationFailure("Regular expression is invalid: missing )")

    if _re:
        # Put `re._pattern_type` or `Regex` object back.
        sub_spec["$regex"] = _re

    if "$options" in new_sub_spec:
        # Remove $options, Monty can't digest it
        del new_sub_spec["$options"]

    return new_sub_spec


"""
Field-level Query Operators
- Comparison
"""


def _is_comparable_ver4(val, qry):
    return gravity(val, weight_only=True) == gravity(
        qry, weight_only=True
    ) or isinstance(qry, (bson.MinKey, bson.MaxKey))


def _is_comparable_ver3(val, qry):
    return gravity(val, weight_only=True) == gravity(qry, weight_only=True)


_is_comparable = _is_comparable_ver4


def _eq_match(fieldwalker, query):
    """ """
    if is_duckument_type(query):
        for val in fieldwalker.value:
            if is_duckument_type(val):
                if compare_documents(query, val):
                    return True

    else:
        if query is None:
            return fieldwalker.value.null_or_missing()

        if isinstance(query, bson.Decimal128):
            query = _cmp_decimal(query)

        for val in fieldwalker.value:
            if isinstance(val, bson.Decimal128):
                val = _cmp_decimal(val)

            if val == query and _is_comparable(val, query):
                return True


def parse_eq(query):
    @keep(query)
    def _eq(fieldwalker):
        return _eq_match(fieldwalker, query)

    return _eq


def parse_ne(query):
    @keep(query)
    def _ne(fieldwalker):
        return not _eq_match(fieldwalker, query)

    return _ne


def parse_gt(query):
    @keep(query)
    def _gt(fieldwalker):
        for value in fieldwalker.value:
            if _is_comparable(value, query):
                if query in bson.decimal128_NaN_ls:
                    return False
                if Weighted(value) > Weighted(query):
                    return True
            elif isinstance(query, (bson.MinKey, bson.MaxKey)):
                return True

    return _gt


def parse_gte(query):
    @keep(query)
    def _gte(fieldwalker):
        for value in fieldwalker.value:
            if _is_comparable(value, query):
                if query in bson.decimal128_NaN_ls:
                    return True if value in bson.decimal128_NaN_ls else False
                if query == bson.decimal128_INF and not value == bson.decimal128_INF:
                    return False
                if Weighted(value) >= Weighted(query):
                    return True
            elif isinstance(query, (bson.MinKey, bson.MaxKey)):
                return True

    return _gte


def parse_lt(query):
    @keep(query)
    def _lt(fieldwalker):
        for value in fieldwalker.value:
            if _is_comparable(value, query):
                if value in bson.decimal128_NaN_ls:
                    return False
                if Weighted(value) < Weighted(query):
                    return True
            elif isinstance(query, (bson.MinKey, bson.MaxKey)):
                return True

    return _lt


def parse_lte(query):
    _dec_NaN_INF_ls = list(bson.decimal128_NaN_ls) + [bson.decimal128_INF]

    @keep(query)
    def _lte(fieldwalker):
        for value in fieldwalker.value:
            if _is_comparable(value, query):
                if query in bson.decimal128_NaN_ls:
                    return True if value in bson.decimal128_NaN_ls else False
                if query == bson.decimal128_INF and value in bson.decimal128_NaN_ls:
                    return False
                if query not in _dec_NaN_INF_ls and value in _dec_NaN_INF_ls:
                    return False
                if Weighted(value) <= Weighted(query):
                    return True
            elif isinstance(query, (bson.MinKey, bson.MaxKey)):
                return True

    return _lte


def _in_match(fieldwalker, query):
    """Helper function for $in and $nin"""
    q_regex = []
    q_value = []
    for q in query:
        if is_pattern_type(q):
            q_regex.append(q)
        elif isinstance(q, bson.Regex):
            try:
                q_regex.append(q.try_compile())
            except re.error as e:
                raise OperationFailure("Regular expression is invalid: {}".format(e))
        else:
            q_value.append(q)

    for q in q_value:
        if _eq_match(fieldwalker, q):
            return True

    for q in q_regex:
        for value in fieldwalker.value:
            if isinstance(value, string_types) and q.search(value):
                return True


def parse_in(query):
    if not isinstance(query, list):
        raise OperationFailure("$in needs an array")

    if any(_is_expression_obj(q) for q in query):
        raise OperationFailure("cannot nest $ under $in")

    @keep(query)
    def _in(fieldwalker):
        return _in_match(fieldwalker, query)

    return _in


def parse_nin(query):
    if not isinstance(query, list):
        raise OperationFailure("$nin needs an array")

    if any(_is_expression_obj(q) for q in query):
        raise OperationFailure("cannot nest $ under $nin")

    @keep(query)
    def _nin(fieldwalker):
        return not _in_match(fieldwalker, query)

    return _nin


"""
Field-level Query Operators
- Array
"""


def parse_all(query):

    field_op_ls = set(QueryFilter({}).field_ops.keys())
    field_op_ls.remove("$eq")
    field_op_ls.remove("$not")

    if not isinstance(query, list):
        raise OperationFailure("$all needs an array")

    if is_duckument_type(query[0]) and next(iter(query[0])) == "$elemMatch":
        go_match = True
        for q in query:
            if not (is_duckument_type(q) and next(iter(q)) == "$elemMatch"):
                raise OperationFailure("$all/$elemMatch has to be consistent")
    else:
        go_match = False
        for q in query:
            if is_duckument_type(q) and next(iter(q)) in field_op_ls:
                raise OperationFailure("no $ expressions in $all")

    @keep(query)
    def _all(fieldwalker):
        if go_match:
            for q in query:
                queryfilter = QueryFilter(q["$elemMatch"])
                doc_type = fieldwalker.doc_type
                for value in fieldwalker.value.iter_arrays():
                    if not any(queryfilter(v, doc_type) for v in value):
                        return False
            return True
        else:
            for q in query:
                if q not in fieldwalker.value:
                    return False
            return True

    return _all


def parse_elemMatch(query):
    # (NOTE) $elemMatch in MontyDB may require document input to proceed
    #        further filter error.OperationFailure check, here we put one
    #        fake doc {}
    QueryFilter(query)({})

    @keep(query)
    def _elemMatch(fieldwalker):
        queryfilter = QueryFilter(query)
        doc_type = fieldwalker.doc_type
        for elem in fieldwalker.value.iter_elements():
            if queryfilter(elem, doc_type):
                return True

    return _elemMatch


def parse_size(query):
    if isinstance(query, float):
        raise OperationFailure("$size must be a whole number")
    if not isinstance(query, int):
        raise OperationFailure("$size needs a number")

    @keep(query)
    def _size(fieldwalker):
        for value in fieldwalker.value.iter_arrays():
            if len(value) == query:
                return True

    return _size


"""
Field-level Query Operators
- Element
"""


def parse_exists(query):
    @keep(query)
    def _exists(fieldwalker):
        return fieldwalker.value.is_exists() == bool(query)

    return _exists


BSON_TYPE_ALIAS_ID = {
    "double": 1,
    "string": 2,
    "object": 3,
    "array": 4,
    "binData": 5,
    # undefined (Deprecated)
    "objectId": 7,
    "bool": 8,
    "date": 9,
    "null": 10,
    "regex": 11,
    # dbPointer (Deprecated)
    "javascript": 13,
    # symbol (Deprecated)
    "javascriptWithScope": 15,
    "int": 16,
    "timestamp": 17,
    "long": 18,
    "decimal": 19,
    "minKey": -1,
    "maxKey": 127,
}


_BSON_TYPE_ID = tuple(BSON_TYPE_ALIAS_ID.values())


def obj_to_bson_type_id(obj):

    BSON_TYPE_ID = {
        float: 1,
        # string: 2,
        bson.SON: 3,
        dict: 3,
        list: 4,
        tuple: 4,
        bson.Binary: 5,
        # bytes: 5,
        # undefined (Deprecated)
        bson.ObjectId: 7,
        bool: 8,
        datetime: 9,
        type(None): 10,
        bson.Regex: 11,
        RE_PATTERN_TYPE: 11,
        # dbPointer (Deprecated)
        # javascript: 13,
        # symbol (Deprecated)
        # javascriptWithScope: 15,
        int: 16,
        bson.Timestamp: 17,
        bson.Int64: 18,
        bson.Decimal128: 19,
        bson.MinKey: -1,
        bson.MaxKey: 127,
    }

    try:
        type_id = BSON_TYPE_ID[type(obj)]
    except KeyError:
        if isinstance(obj, bson.Code):  # also an instance of string_types
            type_id = 13 if obj.scope is None else 15
        elif isinstance(obj, string_types):
            type_id = 2
        elif isinstance(obj, bytes):
            type_id = 5
        elif isinstance(obj, Mapping):
            type_id = 3
        else:
            type_id = None
    finally:
        if type_id is None:
            raise TypeError("Unknown data type: {!r}".format(type(obj)))

        return type_id


def parse_type(query):
    def get_bson_type_id_set(values):
        return set([obj_to_bson_type_id(v) for v in values])

    def str_type_to_int(query):
        if len(query) == 0:
            raise OperationFailure("$type must match at least one type")

        int_types = []
        for q in query:
            if isinstance(q, string_types):
                try:
                    int_types.append(BSON_TYPE_ALIAS_ID[q])
                except KeyError:
                    raise OperationFailure("Unknown type name alias: {}".format(q))
            elif isinstance(q, int):
                if q not in _BSON_TYPE_ID:
                    raise OperationFailure("Invalid numerical type code: {}".format(q))
                int_types.append(q)
            else:
                raise OperationFailure(
                    "type must be represented as a number or a string"
                )
        return int_types

    if not isinstance(query, list):
        query = set(str_type_to_int([query]))
    query = set(str_type_to_int(query))

    @keep(query)
    def _type(fieldwalker):
        if fieldwalker.value.is_exists():
            bids = get_bson_type_id_set(fieldwalker.value)
            return bids.intersection(query)

    return _type


"""
Field-level Query Operators
- Evaluation
"""


def parse_regex(query):
    if isinstance(query, bson.Regex):
        q = query.try_compile()
    else:
        if not isinstance(query["pattern"], string_types):
            raise OperationFailure("$regex has to be a string")
        if not isinstance(query["flags"], (string_types, _FALG)):
            raise OperationFailure("$options has to be a string")

        if isinstance(query["flags"], _FALG):
            flags = query["flags"].retrieve
        else:
            flags = re_str_flags_to_int(query["flags"])

        q = re.compile(query["pattern"], flags)

    @keep(query)
    def _regex(fieldwalker):
        for value in fieldwalker.value:
            if isinstance(value, (string_types, bytes)) and q.search(value):
                return True

    return _regex


def _mod_remainder_not_num_():
    pass


def _mod_remainder_not_num_v42():
    # mongo-4.2.19+
    # https://jira.mongodb.org/browse/SERVER-23664
    raise OperationFailure("malformed mod, remainder not a number")


_mod_remainder_not_num = _mod_remainder_not_num_v42


def parse_mod(query):
    if not isinstance(query, list):
        raise OperationFailure("malformed mod, needs to be an array")
    if len(query) < 2:
        raise OperationFailure("malformed mod, not enough elements")
    if len(query) > 2:
        raise OperationFailure("malformed mod, too many elements")

    divisor = query[0]
    remainder = query[1]

    num_types = (integer_types, float, bson.Decimal128)

    if not isinstance(divisor, num_types):
        raise OperationFailure("malformed mod, divisor not a number")
    if not isinstance(remainder, num_types):
        _mod_remainder_not_num()
        remainder = 0

    if isinstance(divisor, bson.Decimal128):
        divisor = divisor.to_decimal()
    if isinstance(remainder, bson.Decimal128):
        remainder = remainder.to_decimal()

    def mod_scan(field_value, query):
        for value in field_value:
            if isinstance(value, bool) or not isinstance(value, num_types):
                continue
            if isinstance(value, bson.Decimal128):
                value = value.to_decimal()
            if int(value % divisor) == int(remainder):
                return True
        return False

    @keep(query)
    def _mod(fieldwalker):
        field_value = fieldwalker.value
        if mod_scan(field_value, query):
            return True

    return _mod


def parse_jsonSchema(query):
    @keep(query)
    def _jsonSchema(fieldwalker):
        pass

    return _jsonSchema




import os
import lmdb
import shutil
from itertools import islice

from ..types import unicode_, to_bytes, bson
from . import (
    AbstractStorage,
    AbstractDatabase,
    AbstractCollection,
    AbstractCursor,

    StorageDuplicateKeyError,
)


LMDB_DB_EXT = ".mdb"


class LMDBKVEngine(object):
    """Per collection"""

    dbname = to_bytes("documents")

    def __init__(self, options=None):
        """
        """
        opt = options or dict()
        opt.update({
            "subdir": False,
            "max_dbs": 1,
        })
        self.opt = opt
        self._path = None

    def set_path(self, path):
        self._path = path

    def open(self, readonly=None):
        opt = self.opt.copy()
        if readonly is not None:
            opt["readonly"] = readonly

        return lmdb.open(self._path, **opt)

    def iter_docs(self):
        if not os.path.isfile(self._path):
            return

        with self.open(readonly=True) as env:
            db = env.open_db(self.dbname)
            with env.begin(db, write=False) as txn:
                cursor = txn.cursor()
                for encoded_doc in cursor.iternext(keys=False, values=True):
                    yield encoded_doc

    def write(self, environment, pairs, overwrite=False):
        if not os.path.isfile(self._path):
            return

        dup = False
        with environment as env:
            db = env.open_db(self.dbname)
            with env.begin(db, write=True) as txn:
                for doc_id, encoded_doc in pairs:
                    id = bson.id_encode(doc_id)
                    if not txn.put(id, encoded_doc, overwrite=overwrite):
                        dup = True
                        break
        if dup:
            raise StorageDuplicateKeyError()

    def delete(self, environment, doc_ids):
        if not os.path.isfile(self._path):
            return

        with environment as env:
            db = env.open_db(self.dbname, reverse_key=False)
            with env.begin(db, write=True) as txn:
                cursor = txn.cursor()
                for doc_id in doc_ids:
                    id = bson.id_encode(doc_id)
                    if cursor.set_key(id):
                        cursor.delete()


class LMDBStorage(AbstractStorage):
    """
    """

    def __init__(self, repository, storage_config):
        super(LMDBStorage, self).__init__(repository, storage_config)
        self._conn = LMDBKVEngine(self._config)

    def _db_path(self, db_name):
        """
        Get Monty database dir path.
        """
        return os.path.join(self._repository, db_name)

    @classmethod
    def nice_name(cls):
        return "lightning"

    @classmethod
    def config(cls, map_size=10485760, **kwargs):
        """
        """
        return {
            "map_size": int(map_size),
        }

    def database_create(self, db_name):
        if not os.path.isdir(self._db_path(db_name)):
            os.makedirs(self._db_path(db_name))

    def database_drop(self, db_name):
        db_path = self._db_path(db_name)
        if os.path.isdir(db_path):
            shutil.rmtree(db_path)

    def database_list(self):
        return [
            name for name in os.listdir(unicode_(self._repository))
            if os.path.isdir(self._db_path(name))
        ]


class LMDBDatabase(AbstractDatabase):
    """
    """

    def __init__(self, storage, subject):
        super(LMDBDatabase, self).__init__(storage, subject)
        self._db_path = storage._db_path(self._name)
        self._conn = storage._conn

    def _col_path(self, col_name):
        """
        Get database file path, which is Monty collection.
        """
        return os.path.join(self._db_path, col_name) + LMDB_DB_EXT

    def database_exists(self):
        return os.path.isdir(self._db_path)

    def collection_exists(self, col_name):
        return os.path.isfile(self._col_path(col_name))

    def collection_create(self, col_name):
        if not self.database_exists():
            self._storage.database_create(self._name)
        self._conn.set_path(self._col_path(col_name))
        environment = self._conn.open()

        return environment

    def collection_drop(self, col_name):
        if self.collection_exists(col_name):
            os.remove(self._col_path(col_name))

    def collection_list(self):
        if not self.database_exists():
            return []
        return [os.path.splitext(name)[0]
                for name in os.listdir(unicode_(self._db_path))
                if name.endswith(LMDB_DB_EXT)]


LMDBStorage.contractor_cls = LMDBDatabase


class LMDBCollection(AbstractCollection):
    """
    """

    def __init__(self, database, subject):
        super(LMDBCollection, self).__init__(database, subject)
        self._conn = database._conn
        self._conn.set_path(database._col_path(self._name))

    def _ensure_table(func):
        def make_table(self, *args, **kwargs):
            environment = None
            if not self._database.collection_exists(self._name):
                environment = self._database.collection_create(self._name)
            return func(self, env=environment, *args, **kwargs)
        return make_table

    @_ensure_table
    def write_one(self, doc, check_keys=True, env=None):
        env = env or self._conn.open()

        id = doc["_id"]
        encoded = self._encode_doc(doc, check_keys)
        self._conn.write(env, [(id, encoded)])

        return id

    @_ensure_table
    def write_many(self, docs, check_keys=True, ordered=True, env=None):
        env = env or self._conn.open()
        ids = list()

        def produce_encoded_docs():
            for doc in docs:
                id = doc["_id"]
                yield id, self._encode_doc(doc, check_keys)
                ids.append(id)

        self._conn.write(env, produce_encoded_docs())

        return ids

    def update_one(self, doc):
        env = self._conn.open()

        id = doc["_id"]
        encoded = self._encode_doc(doc)
        self._conn.write(env, [(id, encoded)], overwrite=True)

    def update_many(self, docs):
        env = self._conn.open()

        def produce_encoded_docs():
            for doc in docs:
                yield doc["_id"], self._encode_doc(doc)

        self._conn.write(env, produce_encoded_docs(), overwrite=True)

    def delete_one(self, id):
        env = self._conn.open()
        self._conn.delete(env, [id])

    def delete_many(self, ids):
        env = self._conn.open()
        self._conn.delete(env, ids)


LMDBDatabase.contractor_cls = LMDBCollection


class LMDBCursor(AbstractCursor):
    """
    """

    def __init__(self, collection, subject):
        super(LMDBCursor, self).__init__(collection, subject)
        self._conn = self._collection._conn

    def query(self, max_scan):
        docs = (self._decode_doc(doc) for doc in self._conn.iter_docs())

        if not max_scan:
            return docs
        else:
            return islice(docs, max_scan)


LMDBCollection.contractor_cls = LMDBCursor



import os
import contextlib
import importlib
import inspect

from .storage import AbstractStorage, memory
from .errors import ConfigurationError
from .types import string_types


MEMORY_STORAGE = "memory"
SQLITE_STORAGE = "sqlite"
FALTFILE_STORAGE = "flatfile"

DEFAULT_STORAGE = FALTFILE_STORAGE

MEMORY_REPOSITORY = ":memory:"

URI_SCHEME_PREFIX = "montydb://"

MONGO_COMPAT_VERSIONS = ("3.6", "4.0", "4.2", "4.4")  # 4.4 is experimenting


_pinned_repository = {"_": None}
_session = {}
_session_default = {
    "mongo_version": "4.2",
    "use_bson": None,
}
# TODO:
#   The mongo version compating may fail if calling `set_storage()` multiple
#   times with different version.
#   To get this right, may need a factory to spawn CRUD logic object for
#   specific version and hook with client object.
#   Also mind the `MontyClient.server_info`, 'mongoVersion' entry should have
#   the right compat version.


def session_config():
    return _session.copy()


def remove_uri_scheme_prefix(uri_or_dir):
    """Internal function to remove URI scheme prefix from repository path

    Args:
        uri_or_dir (str): Folder path or montydb URI

    Returns:
        str: A repository path without URI scheme prefix

    """
    if uri_or_dir.startswith(URI_SCHEME_PREFIX):
        dirname = uri_or_dir[len(URI_SCHEME_PREFIX):]
    else:
        dirname = uri_or_dir

    return dirname


def provide_repository(dirname=None):
    """Internal function to acquire repository path

    This will pick one repository path in the order of:
    `dirname` -> current pinned repository -> current working dir

    Args:
        dirname (str): Folder path, default None

    Returns:
        str: A repository path acquired from current environment

    """
    if dirname is None or dirname == "":
        return current_repo() or os.getcwd()
    elif isinstance(dirname, string_types):
        return remove_uri_scheme_prefix(dirname)
    else:
        raise TypeError("Repository path should be a string.")


def pin_repo(repository):
    """Pin a db repository for all operations afterward

    Example:
        >>> from montydb import pin_repo, set_storage, MontyClient
        >>> pin_repo("/foo/bar")
        >>> # The following operations will use '/foo/bar'
        >>> set_storage(storage="sqlite")
        >>> client = MontyClient()

    Args:
        repository (str): Database repository path

    """
    _pinned_repository["_"] = provide_repository(repository)


def current_repo():
    """Returns current pinned repository

    Returns:
        str: Database repository path

    """
    return _pinned_repository["_"]


@contextlib.contextmanager
def open_repo(repository=None):
    """Open a repository context

    This will change current working dir to the `repository` or the current
    pinned one during the context. But if the `repository` is ":memory:",
    the current working dir will NOT be changed.

    Args:
        repository (str): Database repository path, default None

    """
    repository = provide_repository(repository)
    crepo = current_repo()
    if repository == MEMORY_REPOSITORY:
        try:
            # Context
            pin_repo(repository)
            yield

        finally:
            pin_repo(crepo)

    else:
        cwd = os.getcwd()
        try:
            # Context
            pin_repo(repository)
            os.chdir(repository)
            yield

        finally:
            pin_repo(crepo)
            os.chdir(cwd)


def find_storage_cls(storage_name):
    """Internal function to find storage engine class

    This function use `importlib.import_module` to find storage module by
    module name. And then it will try to find if there is a class that is
    a subclass of `montydb.storage.abcs.AbstractStorage`.

    Raise `montydb.errors.ConfigurationError` if not found.

    Args:
        storage_name (str): Storage module name

    Returns:
        cls: A subclass of `montydb.storage.abcs.AbstractStorage`

    """
    try:
        monty_storage = "montydb.storage." + storage_name
        module = importlib.import_module(monty_storage)
    except ImportError:
        try:
            module = importlib.import_module(storage_name)
        except ImportError:
            raise ConfigurationError("Storage module '%s' not found." % storage_name)

    for name, cls in inspect.getmembers(module, inspect.isclass):
        if name != "AbstractStorage" and issubclass(cls, AbstractStorage):

            return cls

    raise ConfigurationError(
        "Storage engine class not found. Should "
        "be a subclass of `montydb.storage.abcs."
        "AbstractStorage`."
    )


_storage_ident_fname = ".monty.storage"


def set_storage(
    repository=None, storage=None, mongo_version=None, use_bson=None, **kwargs
):
    """Setup storage engine for the database repository

    Args:
        repository (str, optional): A dir path for database to live on disk.
            Default to current working dir.
        storage (str, optional): Storage module name. Default "flatfile".
        mongo_version (str, optional): Which mongodb version's behavior should
            montydb try to match with. Default "4.2", other versions are "3.6",
            "4.0".
        use_bson (bool, optional): Use bson module. Default `None`.

    keyword args:
        Other keyword args will be parsed as storage config options.

    """
    from .types import bson

    storage = storage or DEFAULT_STORAGE

    if mongo_version and mongo_version not in MONGO_COMPAT_VERSIONS:
        raise ConfigurationError(
            "Unknown mongodb version: %s, currently supported versions are: %s"
            % (mongo_version, ", ".join(MONGO_COMPAT_VERSIONS))
        )

    use_bson = bson.bson_used if use_bson is None else use_bson
    mongo_version = mongo_version or _session.get("mongo_version")

    for key, value in {"use_bson": use_bson, "mongo_version": mongo_version}.items():
        if value is None:
            value = _session_default[key]
        _session[key] = value

    kwargs.update(_session)

    storage_cls = find_storage_cls(storage)

    if storage == MEMORY_STORAGE:
        repository = MEMORY_REPOSITORY
    else:
        repository = provide_repository(repository)
        setup = os.path.join(repository, _storage_ident_fname)

        if not os.path.isdir(repository):
            os.makedirs(repository)

        with open(setup, "w") as fp:
            fp.write(storage)

    storage_cls.save_config(repository, **kwargs)


def provide_storage(repository):
    """Internal function to get storage engine class from config

    Args:
        repository (str): A dir path for database to live on disk.

    Returns:
        A Subclass of `montydb.storage.abcs.AbstractStorage`

    """
    if repository == MEMORY_REPOSITORY:
        storage_name = MEMORY_STORAGE
        if not memory.is_memory_storage_set():
            set_storage(repository, storage_name)

    else:
        setup = os.path.join(repository, _storage_ident_fname)
        if not os.path.isfile(setup):
            set_storage(repository)

        with open(setup, "r") as fp:
            storage_name = fp.readline().strip()

    storage_cls = find_storage_cls(storage_name)
    config = storage_cls.read_config(repository)

    for key in ("use_bson", "mongo_version"):
        _session[key] = config.get(key, _session_default[key])

    _bson_init(_session["use_bson"])
    _mongo_compat(_session["mongo_version"])

    return storage_cls


def _bson_init(use_bson):
    from .types import bson

    if bson.bson_used is None:
        bson.init(use_bson)

    elif bson.bson_used and not use_bson:
        raise ConfigurationError(
            "montydb has been config to use BSON and "
            "cannot be changed in current session."
        )

    elif not bson.bson_used and use_bson:
        raise ConfigurationError(
            "montydb has been config to opt-out BSON and "
            "cannot be changed in current session."
        )

    else:
        # bson.bson_used == use_bson
        pass


def _mongo_compat(version):
    from .engine import queries

    def patch(mod, func, ver_func):
        setattr(mod, func, getattr(mod, ver_func))

    if version.startswith("3"):
        patch(queries, "_is_comparable", "_is_comparable_ver3")
        patch(queries, "_regex_options_check", "_regex_options_")
        patch(queries, "_mod_remainder_not_num", "_mod_remainder_not_num_")

    elif version == "4.0":
        patch(queries, "_is_comparable", "_is_comparable_ver4")
        patch(queries, "_regex_options_check", "_regex_options_")
        patch(queries, "_mod_remainder_not_num", "_mod_remainder_not_num_")

    elif version == "4.2":
        patch(queries, "_is_comparable", "_is_comparable_ver4")
        patch(queries, "_regex_options_check", "_regex_options_v42")
        patch(queries, "_mod_remainder_not_num", "_mod_remainder_not_num_v42")

    else:
        patch(queries, "_is_comparable", "_is_comparable_ver4")
        patch(queries, "_regex_options_check", "_regex_options_")
        patch(queries, "_mod_remainder_not_num", "_mod_remainder_not_num_v42")




import sys

PY3 = sys.version_info[0] == 3


if PY3:
    import collections.abc as abc
    from itertools import zip_longest
    from configparser import ConfigParser

    string_types = str,
    integer_types = int,
    text_type = str

    def iteritems(d, **kw):
        return iter(d.items(**kw))

    unicode_ = (lambda s: s)

    def encode_(s):
        return s

    def decode_(s):
        return s

else:
    import collections as abc
    from itertools import izip_longest as zip_longest
    from ConfigParser import ConfigParser

    string_types = basestring,  # noqa
    integer_types = (int, long)  # noqa
    text_type = unicode  # noqa

    def iteritems(d, **kw):
        return d.iteritems(**kw)

    unicode_ = unicode  # noqa

    def encode_(s):
        return s.encode("utf-8")

    def decode_(s):
        return s.decode(_fs_encoding)
    _fs_encoding = sys.getfilesystemencoding()


__all__ = [
    "ConfigParser",
    "abc",
    "zip_longest",
    "integer_types",
    "string_types",
    "text_type",
    "iteritems",
    "unicode_",
    "encode_",
    "decode_",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['montydb',
 'montydb.engine',
 'montydb.storage',
 'montydb.types',
 'montydb.utils']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'montydb',
    'version': '2.5.0',
    'description': 'Monty, Mongo tinified. MongoDB implemented in Python !',
    'long_description': '\n<img src="artwork/logo.png" alt="drawing" width="600"/>\n\n[![Python package](https://github.com/davidlatwe/montydb/actions/workflows/python-package.yml/badge.svg)](https://github.com/davidlatwe/montydb/actions/workflows/python-package.yml)\n[![Version](http://img.shields.io/pypi/v/montydb.svg?style=flat)](https://pypi.python.org/pypi/montydb)\n[![PyPi downloads](https://img.shields.io/pypi/dm/montydb)](https://pypistats.org/packages/montydb)\n\n> Monty, Mongo tinified. MongoDB implemented in Python!\n\n_Inspired by [TinyDB](https://github.com/msiemens/tinydb) and it\'s extension [TinyMongo](https://github.com/schapman1974/tinymongo)_\n\n\n## What is it?\n\nA pure Python-implemented database that looks and works like [MongoDB](https://www.mongodb.com/).\n\n```python\n>>> from montydb import MontyClient\n\n>>> col = MontyClient(":memory:").db.test\n>>> col.insert_many( [{"stock": "A", "qty": 6}, {"stock": "A", "qty": 2}] )\n>>> cur = col.find( {"stock": "A", "qty": {"$gt": 4}} )\n>>> next(cur)\n{\'_id\': ObjectId(\'5ad34e537e8dd45d9c61a456\'), \'stock\': \'A\', \'qty\': 6}\n```\n\nMost of the CRUD operators have been implemented. You can visit [issue #14](https://github.com/davidlatwe/montydb/issues/14) to see the full list.\n\nThis project is tested against:\n\n- MongoDB: 3.6, 4.0, 4.2 (4.4 on the way💦)\n- Python: 3.7, 3.8, 3.9, 3.10, 3.11\n\n\n## Install\n\n```sh\npip install montydb\n```\n\n- optional, to use *real* `bson` in operation (`pymongo` will be installed)\n    _For minimum requirements, `montydb` ships with it\'s own fork of `ObjectId` in `montydb.types`, so you may ignore this option if `ObjectId` is all you need from `bson`_\n\n    ```sh\n    pip install montydb[bson]\n    ```\n- optional, to use lightning memory-mapped db as storage engine\n    ```sh\n    pip install montydb[lmdb]\n    ```\n\n\n## Storage\n\n🦄 Available storage engines:\n\n* in-memory\n* flat-file\n* sqlite\n* lmdb (lightning memory-mapped db)\n\nDepending on which one you use, you may have to configure the storage engine before you start.\n\n> ⚠️\n>\n> The configuration process only required on repository creation or modification. And, one repository (the parent level of databases) can only assign one storage engine.\n\nTo configure a storage, see flat-file storage for example:\n\n```python\nfrom montydb import set_storage, MontyClient\n\n\nset_storage(\n    # general settings\n    \n    repository="/db/repo",  # dir path for database to live on disk, default is {cwd}\n    storage="flatfile",     # storage name, default "flatfile"\n    mongo_version="4.0",    # try matching behavior with this mongodb version\n    use_bson=False,         # default None, and will import pymongo\'s bson if None or True\n\n    # any other kwargs are storage engine settings.\n    \n    cache_modified=10,       # the only setting that flat-file have\n)\n\n# ready to go\n```\n\nOnce that done, there should be a file named `monty.storage.cfg` saved in your db repository path. It would be `/db/repo` for the above examples.\n\n\n## Configuration\n\nNow let\'s moving on to each storage engine\'s config settings.\n\n### 🌟 In-Memory\n  \n`memory` storage does not need nor have any configuration, nothing saved to disk.\n\n```python\nfrom montydb import MontyClient\n\n\nclient = MontyClient(":memory:")\n\n# ready to go\n```\n\n### 🔰 Flat-File\n  \n`flatfile` is the default on-disk storage engine.\n\n```python\nfrom montydb import set_storage, MontyClient\n\n\nset_storage("/db/repo", cache_modified=5)  # optional step\nclient = MontyClient("/db/repo")  # use current working dir if no path given\n\n# ready to go\n```\n\nFlatFile config:\n\n```ini\n[flatfile]\ncache_modified: 0  # how many document CRUD cached before flush to disk.\n```\n\n### 💎 SQLite\n  \n`sqlite` is NOT the default on-disk storage, need configuration first before getting client.\n\n> Pre-existing sqlite storage file which saved by `montydb<=1.3.0` is not read/writeable after `montydb==2.0.0`.\n\n```python\nfrom montydb import set_storage, MontyClient\n\n\nset_storage("/db/repo", storage="sqlite")  # required, to set sqlite as engine\nclient = MontyClient("/db/repo")\n\n# ready to go\n```\n\nSQLite config:\n\n```ini\n[sqlite]\njournal_mode: WAL\n```\n\nSQLite write concern:\n\n```python\nclient = MontyClient("/db/repo",\n                     synchronous=1,\n                     automatic_index=False,\n                     busy_timeout=5000)\n```\n\n### 🚀 LMDB (Lightning Memory-Mapped Database)\n\n`lightning` is NOT the default on-disk storage, need configuration first before get client.\n\n> Newly implemented.\n\n```python\nfrom montydb import set_storage, MontyClient\n\n\nset_storage("/db/repo", storage="lightning")  # required, to set lightning as engine\nclient = MontyClient("/db/repo")\n\n# ready to go\n```\n\nLMDB config:\n\n```ini\n[lightning]\nmap_size: 10485760  # Maximum size database may grow to.\n```\n\n## URI\n\nOptionally, You could prefix the repository path with montydb URI scheme.\n\n```python\nclient = MontyClient("montydb:///db/repo")\n```\n\n## Utilities\n\n> Pymongo `bson` may required.\n\n* #### `montyimport`\n\n  Imports content from an Extended JSON file into a MontyCollection instance.\n  The JSON file could be generated from `montyexport` or `mongoexport`.\n\n  ```python\n  from montydb import open_repo, utils\n  \n\n  with open_repo("foo/bar"):\n      utils.montyimport("db", "col", "/path/dump.json")\n  \n  ```\n\n* #### `montyexport`\n\n  Produces a JSON export of data stored in a MontyCollection instance.\n  The JSON file could be loaded by `montyimport` or `mongoimport`.\n\n  ```python\n  from montydb import open_repo, utils\n  \n\n  with open_repo("foo/bar"):\n      utils.montyexport("db", "col", "/data/dump.json")\n  \n  ```\n\n* #### `montyrestore`\n\n  Loads a binary database dump into a MontyCollection instance.\n  The BSON file could be generated from `montydump` or `mongodump`.\n\n  ```python\n  from montydb import open_repo, utils\n  \n\n  with open_repo("foo/bar"):\n      utils.montyrestore("db", "col", "/path/dump.bson")\n  \n  ```\n\n* ####  `montydump`\n\n  Creates a binary export from a MontyCollection instance.\n  The BSON file could be loaded by `montyrestore` or `mongorestore`.\n\n  ```python\n  from montydb import open_repo, utils\n  \n\n  with open_repo("foo/bar"):\n      utils.montydump("db", "col", "/data/dump.bson")\n  \n  ```\n\n* #### `MongoQueryRecorder`\n\n  Record MongoDB query results in a period of time.\n  *Requires to access database profiler.*\n\n  This works via filtering the database profile data and reproduce the queries of `find` and `distinct` commands.\n\n  ```python\n  from pymongo import MongoClient\n  from montydb.utils import MongoQueryRecorder\n  \n  client = MongoClient()\n  recorder = MongoQueryRecorder(client["mydb"])\n  recorder.start()\n  \n  # Make some queries or run the App...\n  recorder.stop()\n  recorder.extract()\n  {<collection_1>: [<doc_1>, <doc_2>, ...], ...}\n  \n  ```\n\n* ####  `MontyList`\n\n  Experimental, a subclass of `list`, combined the common CRUD methods from Mongo\'s Collection and Cursor.\n\n  ```python\n  from montydb.utils import MontyList\n  \n  mtl = MontyList([1, 2, {"a": 1}, {"a": 5}, {"a": 8}])\n  mtl.find({"a": {"$gt": 3}})\n  MontyList([{\'a\': 5}, {\'a\': 8}])\n  \n  ```\n\n## Development\n\nmontydb uses [Poetry](https://python-poetry.org/) to make it easy manage dependencies and set up the development environment. \n\n### Initial setup\n\nAfter cloning the repository, you need to run the following commands to set up the development environment:\n\n```bash\nmake install\n```\n\nThis will create a virtual environment and download the required dependencies.\n\n### updating dependencies\n\nTo keep dependencies updated after git operations such as local updates or merging changes into local dev branch\n\n```bash\nmake update\n```\n### Makefile\n\nA makefile is used to simplify common operations such as updating, testing, and deploying etc.\n\n```bash\nmake or make help\n\ninstall                        install all dependencies locally\nupdate                         update project dependencies locally (run after git update)\nci                             Run all checks (codespell, lint, bandit, test)\ntest                           Run tests\nlint                           Run linting with flake8\ncodespell                      Find typos with codespell\nbandit                         Run static security analysis with bandit\nbuild                          Build project using poetry\nclean                          Clean project\n```\n\n### Run mongo docker image\nMost of our tests compare montydb CRUD results against real mongodb instance, therefore we must have a running\nmongodb before testing.\n\nFor example, if we want to test against mongo 4.4:\n```shell\ndocker run --name monty-4.4 -p 30044:27017 -d mongo:4.4\n```\n\n### Tests\n```shell\npoetry run pytest --storage {storage engin name} --mongodb {mongo instance url} [--use-bson]\n```\nExample:\n```shell\npoetry run pytest --storage memory --mongodb localhost:30044 --use-bson\n```\n\n## Why did I make this?\n\nMainly for personal skill practicing and fun.\n\nI work in the VFX industry and some of my production needs (mostly edge-case) requires to run in a limited environment (e.g. outsourced render farms), which may have problem to run or connect a MongoDB instance. And I found this project really helps.\n\n---\n\n<p align=center>\n    <a href="https://jb.gg/OpenSource"><i>This project is supported by JetBrains</i></a>\n</p>\n\n<p align="center">\n    <img src="artwork/icon.png" alt="drawing" width="100"/>\n  &nbsp;&nbsp;\n    <img src="artwork/jetbrains.png" alt="drawing" width="100"/>\n</p>\n',
    'author': 'davidlatwe',
    'author_email': 'davidlatwe@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/davidlatwe/montydb',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<3.12',
}


setup(**setup_kwargs)
